/* 
  DLX Software Simulation

  (C) 1997 by Thomas Schuerger

  E-Mail: schuerge@studcs.uni-sb.de
  Tel:    0681/831406 (Wohnheim D, Uni-Campus)

Version 1_6: 
------------
  06/1997 M.Braun, angepasst fuer HaDePrak, LS Paul
  (mbraun@cs.uni-sb.de)

Version 2_0:
------------
  11/1998 J.Fischer, adaptions for Computer Structure Lab, Tel-Aviv University
  (jofis@cs.uni-sb.de)

Version 2_1:
-----------
  2/99 J.Fischer: bugfix that caused DLXSIM to believe that -1 is greater than 1

Warning:
  Register 0 is writeprotected. Write access to this register is
  just ignored. This preserves zero in the register.

*/

#define VERSION 2
#define REVISION 1

#include <stdio.h>
#include <fcntl.h>

#define RAM_SIZE   0x00100000
#define EPROM_SIZE 0x00010000
#undef  EXT_ALUI

typedef unsigned long ulong;

void InitMem(void);
void FreeMem(void);
int  ReadFile(char *);
void Simulate(void);
void ShowSim(void);
void Step(void);
int StringToHex(char *,unsigned long *);
char *Disassemble(unsigned long);
void ShowHelp(void);
signed long GetNumber(char *);
char *ConvertToSignedHex(signed long);

void  dlx_reset(void);
void  dlx_init(void);
ulong mem_read(ulong);
void  mem_write(ulong,ulong);

char *cycle="\\|/-"; 

int paul=0;         /* Mode: 1 for Paul-mode, 0 for Braun-mode */

#define rs rs1      /* For easier access */

unsigned long *ram=NULL,           /* Pointer to DLX ram (RAM) */
	      *eprom=NULL,         /*   "                (EPROM) */
	       ramsize,            /* DLX RAM size in longwords, power of 2 */
	       steps,              /* Number of steps traced so far */
	       busaccesses;        /* Number of busaccesses so far */

long pc,ir;   /*war ein unsigned*/            /* DLX registers */
int rs1,rs2,rd;                    /* Registers after decode */
int aluf;
signed long imm;                   /* Immediate constant */
long reg[32]; /*war ein unsigned*/            /* 32 32-bit registers */

unsigned long dumpstart=0x800000;  /* Address where hex dumping (RAM)
				      should start on screen.   */

signed long upper=0,lower=15;      /* Dissambler */


void dlx_reset()
{
  pc = 0x0;
  ir = 0x0;

  steps=0;
  busaccesses=0;
}

void dlx_init()
{
  int i; 

  if (eprom!=NULL) 
    for (i=0;i<EPROM_SIZE;i++) eprom[i]=0xffffffff; 
  if (ram!=NULL)
    for (i=0;i<ramsize;i++) ram[i]=0xffffffff; 

  for (i=0;i<32;i++) reg[i]=0x0; 
}

ulong mem_read(adr)
  ulong adr;
{
  if ((adr&0xfff00000)==0x00800000 ) {
    adr = adr & 0x000fffff;
    if (adr<ramsize) return(ram[adr]);
    else             return(0xffffffff);
  }
  else if ((adr&0xffff0000)==0x00000000) {
    adr = adr & 0x0000ffff; 
    return(eprom[adr]);
  }
  return(0xffffffff);
}

void mem_write(adr, val)
  ulong adr; ulong val;
{
  if ((adr&0xfff00000)==0x00800000) {
    adr = adr & 0x000fffff;
    if (adr<ramsize) ram[adr]=val; 
  }
  else if ((adr&0xffff0000)==0x00000000) {
    adr = adr & 0x0000ffff; 
    eprom[adr]=val;
  }
}

void usage() 
{
  printf("Usage: dlxsim [-[h]] [.COD-File]\n");
  exit(0);
}

void main(int argc,char **argv)
{

  printf("\n* DLX Processor Simulation Software V%lu.%lu *\n\n",VERSION,REVISION);
  printf("  A tool for the Computer Structure Lab by Dr. Guy Even\n"); 
  printf("  Tel-Aviv University\n\n");
  printf("  -> Originalcode (C) 1997 by Thomas Schuerger <-\n");
  printf("     Extensions(V1.6), 06/1997, M.Braun\n"); 
  printf("     Translation(V2.0), 11/1998, J.Fischer\n"); 
  printf("     Bugfix(V2.1), 2/99, J.Fischer\n");
// printf("\n(Email: schuerge@studcs.uni-sb.de, mbraun@cs.uni-sb.de)\n"); 
  printf("\n");
  
  if ((argc>2)||((argc==2)&&(argv[1][0]=='-'))) usage(); 

  InitMem();
  dlx_reset();
  dlx_init(); 

  ReadFile("eprom.cod");  

  if (argc==2) ReadFile(argv[1]);

  printf("\nPress Return to continue (h for help)...."); fflush(stdout); 
  while(getch()!=0x0d) ;

  Simulate(); 

  FreeMem(); 
}

/* This function dynamically allocates DLX simulation memory. It
   tries to get 8 MB at maximum, and if this is not available
   it halves the value until the memory can be allocated.        */

void InitMem(void) 
{
  int a;
  unsigned long *ptr;

  eprom = (unsigned long*)malloc(EPROM_SIZE*sizeof(unsigned long)) ; 

  ramsize=RAM_SIZE;  
  while(!(ram=(unsigned long *)malloc(ramsize*sizeof(unsigned long))))
  {
    ramsize=ramsize>>1;
    
    if(ramsize<1024) break; 
  }
  if ( (eprom==NULL)||(ram==NULL) ) {
    printf("Not enough free EPROM/RAM for simulation.\n");
    exit(0);
  }
  printf("Allocated %ld longwords (%ld KB) for simulation EPROM.\n",EPROM_SIZE, EPROM_SIZE>>8);
  printf("Allocated %ld longwords (%ld KB) for simulation RAM.\n",ramsize,ramsize>>8);
}

void FreeMem()
{
  if (eprom!=NULL) {
    free(eprom); eprom=NULL; 
  }
  if (ram!=NULL) {
    free(ram); ram=NULL; 
  }
}

/* The following routine reads the DLX assembler output file into
   memory with all features including memory filling with .DS etc. */

int ReadFile(char *name)
  {
  FILE *f;
  char buf[256];
  int  i;
  int  baseset=0;
  unsigned long cod,mem;
  unsigned long value;
  unsigned long codesize=0, codebase=0;          

  if(f=fopen(name,"r"))
    {
    while(!feof(f))
      {
      if(!fgets(buf,255,f))   /* No more lines to read -> quit loop */
	break;

      if((!strncmp(buf,".CODE",5)) || (!strncmp(buf,".DATA",5)))
	{
	fgets(buf,255,f);
	StringToHex(buf,&mem);
	if(!baseset)
	  {
	  codebase=mem;
	  baseset=1;
	  }
	fgets(buf,255,f);
	StringToHex(buf,&cod);
	codesize+=cod;

	for(i=0;i<cod;i++)
	  {
	  fgets(buf,255,f);
	  StringToHex(buf,&value);
	  mem_write(mem++,value);
	  }
	}
      else if(!strncmp(buf,".DS",3))
	{
	fgets(buf,255,f);
	StringToHex(buf,&mem);
	fgets(buf,255,f);
	StringToHex(buf,&cod);
	codesize+=cod;
	fgets(buf,255,f);
	StringToHex(buf,&value);
	for(i=0;i<cod;i++)
	  mem_write(mem+i,value);
	}
      else if(buf[0] && (buf[0]!='\n' && buf[0]!='\r'))
	{
	printf("File <%s> not a CODE file.\n",name);
	fclose(f);
        return(0); 
	}
      }
    printf("File <%s> read. Baseaddress: $%08lx, Codesize: %lu longword%s\n",name,codebase,codesize,codesize!=1 ? "s" : "");
    fclose(f);
    return(1);
    }
  else
    {
    printf("Cannot open file \"%s\".\n",name);
    return(0);
    }
  return(0);
}

/* Main simulation routines */

void Simulate()
  {
  int valid=0,quit=0;

  signed long a,b,c;
  signed short immbuf;
  int i;
  char string[256],buf[256];
  char *ptr;
  unsigned long m,eadr;
  signed long arg;

  dlx_reset(); 

  upper=pc-7; lower=pc+8; 
  printf("\n");

  while(!quit)
  {
    clrscr(); //jofis
    ShowSim();

    valid=0;
    printf("\n"); 
    printf(" Command: ");fflush(stdout);

    do
    {
       printf("");fflush(stdout);
       gets(string);
       ptr=string;   

      if(string[0]==0)
      {
	Step();
	valid=1;
      }
      else if(string[0]=='?' || string[0]=='h')
      {
        clrscr(); //jofis
	ShowHelp();
	valid=1;
      } 
      else if (strncmp(ptr,"reset",5)==0) {
        dlx_reset();
        upper = pc-7; lower = pc+8;
        valid = 1; 
      }
      else if (strncmp(ptr,"init",4)==0) {
        dlx_init();
        valid = 1; 
      }
      else if ((string[0]=='q')||((strncmp(ptr,"quit",4)==0)))  
      {
	quit=1;
	valid=1;
      } 
      else if(string[0]=='c')
      {
	steps=0;
	busaccesses=0;
	valid=1;
      }
      else if (strncmp(ptr,"l ",2)==0) {
        if (*(ptr+2)!=0) {
	  printf("\r");fflush(stdout);
          ReadFile(ptr+2);
          { 
            printf("Press Return to continue...."); fflush(stdout); 
            while(getch()!=0x0d) ;
          }
        }
        valid=1; 
      }
      else if(string[0]=='f')
      {
	a=GetNumber(ptr+1);
	printf("\rStart address: ");fflush(stdout);
	gets(string);
	if(string[0])
	  {
	  b=GetNumber(string);
	  printf("\r  End address (use \"+<n>\" = number of bytes): ");fflush(stdout);
	  gets(string);
	  if(string[0])
	  {
	    if(string[0]=='+')
	    {
	      c=GetNumber(string+1);
	      m   =b;
	      eadr=m+c;
	      while(m<eadr) {
		mem_write(m++,a);
	      }
	    }
	    else
	      {
	      c=GetNumber(string);
	      if(c>=b)        
	      {
		m=b;
		eadr=c;
		while(m<=eadr)
		{
		  mem_write(m++,a);
		}
	      }
	    }
	  }
	}
	valid=1;
      }
      else if(string[0]=='s')
      {
	if (string[1]!=0)
          if ((string[2]=='R')||(string[2]=='r')) 
            a=GetNumber(ptr+3);
          else a=GetNumber(ptr+1); 
        else a=0;
        if ((0<a)&&(a<32)) { 
	  printf("\rValue: ");fflush(stdout);
	  gets(string);
	  if(string[0])
	  {
	    b=GetNumber(string);
            reg[a]=b; 
          }
        }
	valid=1;
      }
      else if(string[0]=='g')
      {
	a=GetNumber(ptr+1);
	printf("\r");fflush(stdout);
	printf("  : Running until PC=$%08lx... Press any key to stop.",a);fflush(stdout);

	while ((pc!=a)&&(!kbhit())) {
          Step();
          if ((steps&0xffff)==0x0000) 
            printf("\r%c",cycle[(steps>>16)&0x3]); fflush(stdout);
        }

	printf("");fflush(stdout);
	valid=1;
      }       
      else if(!strncmp(string,"r",1))
      {
	a=GetNumber(ptr+1);
	if(a>0)
        {
	  printf("\r");fflush(stdout);
	  printf("  : Running %d steps (fast)... Press any key to stop.",a);fflush(stdout);
	  for(i=0;((i<a)&&(!kbhit()));i++) {
	    Step();
            if ((steps&0xffff)==0x0000) 
              printf("\r%c",cycle[(steps>>16)&0x3]); fflush(stdout);
          }
	  valid=1;
        }
      }
      else if (!strncmp(string,"n",1))
      { 
	a=GetNumber(ptr+1);
	if(a>0)
	{
	  printf("\r");fflush(stdout);
	  printf("  : Running %d steps... Press any key to stop.",a);fflush(stdout);
	  for(i=0;((i<a)&&(!kbhit()));i++) {
	    Step();
            clrscr(); //jofis
	    ShowSim();
            if(((signed long)pc<=upper) || ((signed long)pc>=lower))
            {    
              /* Center PC vertically on screen */
              upper=pc-7; lower=pc+8;
            }
            printf("\n%c",cycle[(steps>>2)&0x3]); fflush(stdout);
          }
	  valid=1;
	}
      }
      else if(!strncmp(string,"m",1))
      {
	paul^=1;
	valid=1; 
      }
      else if(!strncmp(string,"p",1))
      {
	a=GetNumber(ptr+1);
	if(a<0)
	  printf("Conversion: %ld = %s ($%08lx)",a,ConvertToSignedHex(a),a);
	else
	  printf("Conversion: %ld = %s ($%08lx)",a,ConvertToSignedHex(a),a);

	fflush(stdout);
        while(getch()!=0x0d);
	valid=1;
      }
      else if(!strncmp(string,"j",1))
      {
	pc=GetNumber(ptr+1);
	valid=1;
      }
      else if(!strncmp(string,"d",1))
      {
	dumpstart=GetNumber(ptr+1);
	valid=1;
      }
    } while(!valid);

    if(((signed long)pc<=upper) || ((signed long)pc>=lower))
    {    
      /* Center PC vertically on screen */
      upper=pc-7; lower=pc+8;
    }
  }
}

void ShowSim(void) 
{
  int    i; 
  signed long a;

  printf("");   

  for(i=0,a=(signed long)upper;a<=lower;a++,i++)
  {
    if(a==pc)
      printf("$%08lx:  > %-24s R%02lu = $%08lx    R%02lu = $%08lx\n",a,Disassemble(a),i,reg[i],i+16,reg[i+16]);
    else
      printf("$%08lx:    %-24s R%02lu = $%08lx    R%02lu = $%08lx\n",a,Disassemble(a),i,reg[i],i+16,reg[i+16]);
  }
  printf("                                                                            \n");
  for(i=0;i<5;i++)
  {
    switch(i)
    {
      case 0: printf("      PC: $%08lx    $%08lx: $%08lx $%08lx $%08lx $%08lx\n",pc,dumpstart+i*4,mem_read(dumpstart+i*4),mem_read(dumpstart+i*4+1),mem_read(dumpstart+i*4+2),mem_read(dumpstart+i*4+3));break;
      case 1: printf("  PC Cmd: $%08lx    $%08lx: $%08lx $%08lx $%08lx $%08lx\n",mem_read(pc),dumpstart+i*4,mem_read(dumpstart+i*4),mem_read(dumpstart+i*4+1),mem_read(dumpstart+i*4+2),mem_read(dumpstart+i*4+3));break;
      case 2: printf("Executed: %-10lu   $%08lx: $%08lx $%08lx $%08lx $%08lx\n",steps,dumpstart+i*4,mem_read(dumpstart+i*4),mem_read(dumpstart+i*4+1),mem_read(dumpstart+i*4+2),mem_read(dumpstart+i*4+3));break;
      case 3: printf("Bus XSes: %-10lu   $%08lx: $%08lx $%08lx $%08lx $%08lx\n",busaccesses,dumpstart+i*4,mem_read(dumpstart+i*4),mem_read(dumpstart+i*4+1),mem_read(dumpstart+i*4+2),mem_read(dumpstart+i*4+3));break;
      default: printf("    Mode: %-10s   $%08lx: $%08lx $%08lx $%08lx $%08lx\n",paul ? "$imm(Reg)" : "Reg,$imm",dumpstart+i*4,mem_read(dumpstart+i*4),mem_read(dumpstart+i*4+1),mem_read(dumpstart+i*4+2),mem_read(dumpstart+i*4+3));break;
    }
  }
}

void Step(void)
  {
  signed short immbuf;
  unsigned int a;

  ir=mem_read(pc++);   /* Fetch */
  steps++;
  busaccesses++;

  if((ir>>26)==0)   /* RTYPE command? */
    {
    rs1=(ir>>21)&31;
    rs2=(ir>>16)&31;
    rd=(ir>>11)&31;
    aluf=ir & 0x3f;
    }
  else
    {
    immbuf=ir&65535;           /* Extract immediate value */
    imm=(signed long)immbuf;   /* Sign extend immediate */
    rd =(ir>>16)&31;           /* Extract register destination */
    rs1=(ir>>21)&31;           /* Extract register source */
    }

  switch(a=(ir>>26))
    {
    case 35:   /* Load */

	       if(rd!=0)
		 reg[rd]=mem_read(reg[rs1]+imm);
	       busaccesses++;
	       break;

    case 43:   /* Store */

	       mem_write(imm+reg[rs1],reg[rd]);               
	       busaccesses++;
	       break;

      case  4: /* BEQZ */

	       if(reg[rs1]==0) pc+=imm;
	       break;

      case  5: /* BNEZ */

	       if(reg[rs1]!=0) pc+=imm;
	       break;
      case 11: /* ADDI */
	       if(rd!=0) reg[rd]=reg[rs1]+imm;
	       break;
#ifdef EXT_ALUI 
      case 10: /* SUBI */
  	       if(rd!=0) reg[rd]=reg[rs1]-imm;
               break;
      case 14: /* ANDI */
               if(rd!=0) reg[rd]=reg[rs1]&imm;
               break;
      case 13: /* ORI */
               if(rd!=0) reg[rd]=reg[rs1]|imm;
               break;
      case 12: /* XORI */
               if(rd!=0) reg[rd]=reg[rs1]^imm;
               break;
#endif
      case 23: /* JALR */
	       reg[31]=pc;
	       pc=reg[rs];
	       break;
      case 22: /* JR */
	       pc=reg[rs];
	       break;

      case  0: /* ALU */
	       switch(aluf)
		 {
		 case 32+3: reg[rd]=reg[rs1]+reg[rs2];break;
		 case 32+2: reg[rd]=reg[rs1]-reg[rs2];break;
		 case 32+6: reg[rd]=reg[rs1]&reg[rs2];break;
		 case 32+5: reg[rd]=reg[rs1]|reg[rs2];break;
		 case 32+4: reg[rd]=reg[rs1]^reg[rs2];break;
		 case 0   : reg[rd]=(reg[rs1]<<1)&0xfffffffe;break;
		 case 2   : reg[rd]=(reg[rs1]>>1)&0x7fffffff;break;
		 }
	     break;
      default:
	if((a>>3) == 3 && rd!=0) /* sreli */
	  {
	  switch(a&7)
	    {
	    case 1: reg[rd]=(reg[rs1]>imm) ? 1 : 0;break;
	    case 2: reg[rd]=(reg[rs1]==imm) ? 1 : 0;break;
	    case 3: reg[rd]=(reg[rs1]>=imm) ? 1 : 0;break;
	    case 4: reg[rd]=(reg[rs1]<imm) ? 1 : 0;break;
	    case 5: reg[rd]=(reg[rs1]!=imm) ? 1 : 0;break;
	    case 6: reg[rd]=(reg[rs1]<=imm) ? 1 : 0;break;
	    }
	  }
      }
   
}

/* This routines disassembles a memory location and returns
   a string with the disassembled command. Note that you should
   copy the string to another one if you want to use it several times.
   Using printf() with the returned value is okay.                     */

char *Disassemble(unsigned long adr)
  {
  unsigned int a;
  unsigned long cmd;
  static char string[150];
  signed short immbuf;
  char *immstr; 

  ir=mem_read(adr);   /* Fetch */

  if((ir>>26)==0)   /* RTYPE command? */
    {
    rs1=(ir>>21)&31;
    rs2=(ir>>16)&31;
    rd=(ir>>11)&31;
    aluf=ir & 0x3f;
    }
  else
    {
    immbuf=ir&65535;           /* Extract immediate value */
    imm=(signed long)immbuf;   /* Sign extend immediate */
    rd =(ir>>16)&31;           /* Extract register destination */
    rs1=(ir>>21)&31;           /* Extract register source */     
    }

  immstr=ConvertToSignedHex(imm);


  switch(a=(ir>>26))
    {
    case 35: /* LW */
	     if(paul) sprintf(string,"lw     R%lu,%s(R%lu)",rd,immstr,rs);
	     else     sprintf(string,"lw     R%lu,R%lu,%s",rd,rs,immstr);
	     break;
    case 43: /* SW */
	     if(paul) sprintf(string,"sw     %s(R%lu),R%lu",immstr,rs,rd);
	     else     sprintf(string,"sw     R%lu,R%lu,%s",rd,rs,immstr);
	     break;
    case  4: /* BEQZ */
	     sprintf(string,"beqz   R%lu,%s",rs,immstr);
	     break;
    case  5: /* BNEZ */
	     sprintf(string,"bnez   R%lu,%s",rs,immstr);
	     break;
    case 11: /* ADDI */
	     sprintf(string,"addi   R%lu,R%lu,%s",rd,rs,immstr);
	     break;
#ifdef EXT_ALUI
    case 10: /* SUBI */
  	     sprintf(string,"subi   R%lu,R%lu,%s",rd,rs,immstr);
  	     break;
    case 14: /* ANDI */
  	     sprintf(string,"andi   R%lu,R%lu,%s",rd,rs,immstr);
  	     break;
    case 13: /* ORI */
  	     sprintf(string,"ori    R%lu,R%lu,%s",rd,rs,immstr);
  	     break;
    case 12: /* XORI */
  	     sprintf(string,"xori   R%lu,R%lu,%s",rd,rs,immstr);
  	     break;
#endif
    case 23: /* JALR */
	     sprintf(string,"jalr   R%lu",rs);
	     break;
    case 22: /* JR */
	     sprintf(string,"jr     R%lu",rs);
	     break;
    case  0: 
	     switch(aluf)
	     {
		 case 32+3: 
			sprintf(string,"add    R%lu,R%lu,R%lu",rd,rs1,rs2);
			break;
		 case 32+2: 
			sprintf(string,"sub    R%lu,R%lu,R%lu",rd,rs1,rs2);
			break;
		 case 32+6: 
			sprintf(string,"and    R%lu,R%lu,R%lu",rd,rs1,rs2);
			break;
		 case 32+5: 
			sprintf(string,"or     R%lu,R%lu,R%lu",rd,rs1,rs2);
			break;
		 case 32+4: 
			sprintf(string,"xor    R%lu,R%lu,R%lu",rd,rs1,rs2);
			break;
		 case 0: 
			sprintf(string,"slli   R%lu,R%lu",rd,rs1);
			break;
		 case 2: 
			sprintf(string,"srli   R%lu,R%lu",rd,rs1);
			break;
		 default:
			sprintf(string,"????   ($%08lx)",ir);
	      }
	      break;

      case 48: /* Special NOP */
	       sprintf(string,"nop    ($%08lx)",ir); 
	       break;

      default: if((a>>3)==3)   /* Sreli */
		 {
		 switch(a&7)
		 {
		   case 1: 
			sprintf(string,"sgti   R%lu,R%lu,%s",rd,rs,immstr);
			break;
		   case 2: 
			sprintf(string,"seqi   R%lu,R%lu,%s",rd,rs,immstr);
			break;
		   case 3:
			sprintf(string,"sgei   R%lu,R%lu,%s",rd,rs,immstr);
			break;
		   case 4:
			sprintf(string,"slti   R%lu,R%lu,%s",rd,rs,immstr);
			break;
		   case 5:
			sprintf(string,"snei   R%lu,R%lu,%s",rd,rs,immstr);
			break;
		   case 6: 
			sprintf(string,"slei   R%lu,R%lu,%s",rd,rs,immstr);
			break;
		   default:
			sprintf(string,"????   ($%08lx)",ir);
		 }
	       }
	       else sprintf(string,"????   ($%08lx)",ir);break;
      }

  return(string);
  }

/* This routine converts a string from the assembler output file
   into a number that is stored in dest. Only hexadecimal and
   binary stuff is supported.                                    */

int StringToHex(char *was,unsigned long *dest)
  {
  char *ptr;
  int len=0;
  unsigned long sh,num=0;

  if(!strncmp(was,"0x",2)) was+=2;
  if((*was)=='$') was++;
  if((*was)=='%')
    {
    was++;
    ptr=was;
    while(isdigit(*ptr++)) len++;
    
    sh=1<<(len-1);
    while(sh!=0)
      {
      if((*was++)!='0') num+=sh;
      sh>>=1;
      }  
    *dest=num;
  
    return(1);
    }
  sscanf(was,"%x",dest);
  return(1);
  }

/* GetNumber() converts a string into a signed number allowing
   stuff like hexadecimal, binary and decimal conversion.
   Possible strings would be
     "155"
     "$154",
     "-0x154",
     "%1010101"
     etc.                                                      */

signed long GetNumber(char *was)
  {
  unsigned long sh,num=0;
  char *ptr;
  int len=0;
  int sign=0;

  if(isspace(*was)) was++;

  while(*was=='-')
    {
    was++;
    sign^=1;
    }

  if(*was=='%')
    {
    was++;
    ptr=was;
    while(isdigit(*ptr++)) len++;
    
    sh=1<<(len-1);
    while(sh!=0)
      {
      if((*was++)!='0') num+=sh;
      sh>>=1;
      }  
    }
  else if(was[0]=='0' && was[1]=='x')
    {
    was+=2;
    sscanf(was,"%x",&num);
    }
  else if(*was=='$')
    {
    was++;
    sscanf(was,"%x",&num);      
    }
  else
    {
    num=atol(was);
    }
  return(sign ? -num : num);
  }

char *ConvertToSignedHex(signed long a)
{
  static char string[30];

  sprintf(string,"$%lx",a&0xffff);
  
  return(string);
}

void ShowHelp()
{
  printf("DLX Simulator V%lu.%lu Help\n-----------------------\n",VERSION,REVISION);
  
  printf("<CR>    Step     Execute 1 step.\n\n");
  printf("n <n>   Next     Execute n steps.\n");
  printf("r <n>   Run      Execute n steps (fast).\n");
  printf("g <n>   Go To    Run until PC = n.\n");
  printf("j <n>   Jump     Force jump to location n.\n\n");

  printf("reset   Reset    (pc=0x00000000).\n");
  printf("init    Init     Initialize memory, register.\n");  
  printf("l file  Load     Load <file> to memory (cod-format).\n");
  printf("f <n>   Fill     Fill memory with value n. You are asked for start and\n");
  printf("                 end address. Begin end address with \"+\" to specify length.\n");
  printf("s <rd>  Set      Set Register <rd> to a value.\n\n");
  
  printf("p <n>   Print    Print n in decimal and hex.\n");
  printf("d <n>   Dump     Set memory dump location to n.\n");
  printf("m       Mode     Switch between disassemble modes.\n");
  printf("c       Clear    Reset executed steps and bus accesses back to 0.\n\n");

  printf("q/quit  Quit     Free resources and quit.\n");
  printf("-------------------------------------------------------\n");
  printf("Numbers may be specified in the international format:\n");
  printf("Start with $, 0x or %% for hexadecimal or binary numbers.");
  fflush(stdout);

  while(getch()!=0x0d) ;
  printf("");
}




	     

